; GLADOS - Galaxy Line Analysis for Detection of Sources
; BRIEF SUMMARY
; This is the main program, run this first. Full usage guide below.

; HI Source extraction program. Examines average cube and individual polarisations,
; checking for sources above specified S/N and width minimiser thresholds.
; Checks spectra at every pixel, integrating over 3x3 area
; Calls starlink "tmatch" packages to match detections in adjacent pixels

; In case the matching procedure over-merges seperate sources that are close in space 
; and/or velocity, the original output is retained in 'Output.txt'. The main catalogue 
; is found in 'Catalogue.txt', while 'Catalogue.eps' shows the actual spectra after
; the merging.

; *Should* run on any set of cubes, but they must have identical coordinate systems,
; i.e. the pixel position x,y,z in cube A must correspond to the same spatial and 
; velocity position at pixel x,y,z in cube B. Probably it's possible to correct for
; coordinate differences, but I don't want to.

; KNOWN ISSUES :
; Program is designed to reproduce the miriad functions found in "mbspect", but often
; gives slightly different results. In particular the estimated S/N is slightly lower 
; than in miriad, particularly for strong sources. Not so much a problem for faint ones.
; Velocity width estimate is also lower than that from miriad.
; The program has developed a nasty habit of failing to call the starlink packages 
; necessary for the final stages. This causes GLADOS to crash, but the problems can be
; circumvented by afterwards running "CHELL" and then "WHEATLEY" - see usage guide.

; DEVELOPMENT NOTES
; This version constrains the spectra in pol A and B to only be searched along the
; velocity range determined from the detection in C (the averaged or "combined" cube)
; This version lets you set the flux level at which to estimate the velocity width, but
; this is probably pointless - default is to use 50% of peak flux and this works well
; This version does 2 sigma clipping, 5 iterations to estimate rms and S/N
; This version attempts to integrate the spectra and weight by beam parameters in 
; exactly the same way that miriad does
; Works in Cardiff. Probably works everywhere else, too.

; USAGE GUIDE
; 1) First run "chmod +x CHELL.cmd" if necessary.
;
; 2) Set thresholds, channel range and cube names (include full path) below.
; It is recommended that the averaged cube have standard hanning smoothing, width 3,
; applied. The individual polarisations may require additional smoothing (I use width 5
; in addition to the width 3 already applied).
; The velocity range searched (channels z1 to z2) should avoid regions of RFI. 
; Typically this means searching channel ranges spanning the equivalent to 3,000-5,000 
; km/s of velocity, so it's necessary to search about 5 different chunks of the cube
; to search the whole thing. Increasing the velocity range that each chunk spans is not
; recommended, since baseline ripple becomes harder to remove (the program applies a 
; first-order correction). 
; The velocity range masked (vtol) is applied to individual spectra, centered on the 
; peak flux. This prevents detected sources from contributing to the rms measurement
; which would screw up the S/N estimate. Since sources are typically not wider than
; 600 km/s, a vtol of 1,200 km/s ensures that if a peak at the edge of a double-horn
; profile is detected, the whole source will still be masked. It may be necessary to
; lower vtol if the velocity range searched is small.
; The parameter "vw" is the required minimised velocity width a source must have to be
; accepted as a candidate. This is the smallest seperation of points on the spectrum
; at 50% of the peak flux, within the velocity range determined from vtol. It is not
; a great way to estimate a source's true velocity width, but it's pretty good as a way
; to determine if a source is resolved in velocity or not.
;
; 2) Alter any parameters for source matching in "CHELL.cmd". There are two relevant
; keywords with settings which may need adjusting. The default "params='3 3 27'" 
; instructs starlink to search in 3x3 pixel boxes of length 27 channels (about 150 km/s)
; for any other detections. "GroupSize>=3" means that any isolated pixels, or pairs of
; pixels, are removed. Increasing the value of 3 can remove even more spurious 
; detections, but at the cost of completeness.
; IMPORTANT : "CHELL.cmd" is just a script that uses starlink packages. Different 
; computer networks have different requirements to call these - i.e. at Arecibo it is
; necessary for the user to first run "starlink", in Cardiff this is not required. The
; network may or may not require the full path to these packages, this version uses the
; path for Cardiff.
; To alter the path name through which GLADOS finds CHELL, search for "CHELL.cmd" - it
; appears on only one line in the program.
;
; 3) To actually run the the program the user should only need to type ".run GLADOS.pro"
; within an IDL terminal. Because of the network differences mentioned above, the 
; programs "CHELL.cmd" and "WHEATLEY.pro" can also be run independently if GLADOS has 
; problems calling CHELL (ordinarily this is not necessary). In this case, after running 
; GLADOS, run CHELL from a standard xterm, and then WHEATLEY in an IDL prompt. This 
; also means that the user can alter the matching parameters afterwards, as the raw 
; output is preserved (see below).
;
; 4) The positions of possible sources are listed in "Output.txt". The columns are
; labelled in the file. The last 3 are vmin (the minimum width of the source), zm1 and
; zm2 (the channel range the program masks when fitting the baseline and calculating rms).
; Since sources will typically be detected in at least 3 adjacent pixels (or tens of
; pixels for resolved sources) the file "Catalogue.txt" contains only the highest S/N
; detection from each source. This can often combine sources, hence the raw output is
; retained. I recommend plotting the x,y positions in "Output.txt" in Topcat and
; overlaying the "Catalogue.txt" positions.
; The main file for use extracting sources is "Catalogue.eps", which is a postscript
; file of all the spectra at the positions listed in "Catalogue.txt". The region masked
; is indicated by two solid lines, the center of which corresponds to the purported
; detection.
;
; Finally, the program is designed specifically for AGES data and will likely require
; lots of adjustments for other data sets. It's best to run on a machine with plenty of
; RAM as it has to load 3 FITS files into memory. It takes about 15 minutes to run over
; 5,000 km/s of baseline on a 20 square degree cube on the herschel01 machine. The code
; is not very elegant but it does have lots and lots of comments.

; START OF THE BIT THE USER SHOULD PAY ATTENTION TO
;Thresholds
S = 4.0 	; Primary S/N threshold
Si= 3.5         ; S/N threshold in individual polarisations
vtol=1200.0     ; Velocity range to mask, km/s.
vw = 20.0       ; Required minimum velocity width, km/s.
z1 = 2855       ; Minimum channel
z2 = 3900       ; Maximum channel

sizez = z2 - z1
value = dblarr(sizez)

; Files to operate on
averagecube = string('/home/virgo/spxrt/AGES7448Stuff/AGES7448_h3.fits')
polAcube = string('/home/virgo/spxrt/AGES7448Stuff/AGES7448A_h3_5.fits')
polBcube = string('/home/virgo/spxrt/AGES7448Stuff/AGES7448B_h3_5.fits')

; Velocity width measures
WL = 2.0		; Where to measure the velocity width. 2 = 50% of peak flux, 3 = 33% etc.
Wfrac = 1.0/WL
; END OF BIT THE USER SHOULD PAY ANY ATTENTION TO

; Declare variable for counting things. L for long integer type
n = 0L		; Total number of spectra examined
n1= 0L		; Number of matches in C cube
n2= 0L		; Number of matches in C and A cubes
n3= 0L          ; Number of matches in C, A and B cubes
nv= 0L		; Number of valid spectra with no NaN errors
ng= 0L          ; Number of valid spectra in current beam area
width = 0.0	; Velocity width measured
SW = 1.0	; Sum of weights when integrating over beam


; Print Topcat-friendly header to the 'Output.txt' file
openw, lun, 'Output.txt', /get_lun
printf, lun, '#      x       y	   z	     RA	        Dec	   velocity	      SN	    vmin     zm1     zm2'
free_lun, lun

; Set and open the catalogue file for postscript printing
set_plot, 'PS'
device, filename='Catalogue.eps'
!p.multi = [0,2,5]

print, 'Reading in average cube...'
fxread, averagecube, DATAC, HEADERC
print, 'Reading in polarisation A...'
fxread, polAcube, DATAA, HEADERA
print, 'Reading in polarisation B...'
fxread, polBcube, DATAB, HEADERB

print, 'Sanity checks :'
ndims = FXPAR(HEADERC,'NAXIS')
print, 'Number of axes :', ndims
sizex = fxpar(headerc, 'NAXIS1')
sizey = fxpar(headerc, 'NAXIS2')
fullsizez = fxpar(headerc, 'NAXIS3')
print, 'X, y, z sizes :', sizex, sizey, fullsizez
restfq = fxpar(headerC, 'RESTFREQ')
print, 'Rest frequency :', restfq
beams = fxpar(headerc, 'BMAJ')
print, 'Beam size in arcminutes :', beams*60.0
beams2 = fxpar(headerc, 'BMIN')
if (beams ne beams2) then begin
   print, 'WARNING  - non circular beam, using major axis to proceed...'
endif
cdelt1 = fxpar(headerc, 'CDELT1')
cdelt2 = fxpar(headerc, 'CDELT2')
print, 'X-pixel size, arcminutes :', cdelt1*60.0
print, 'Y-pixel size, arcminutes :', cdelt2*60.0

; Get the velocity resolution and use this to set tolerance values
deltav = fxpar(headerc, 'CDELT3')/(-1000.0)
ctol = fix(vtol/deltav) 
vwctol = fix(vw/deltav)

; Parameters to convert channels to velocity
ls = 299792458D			; Lightspeed
zref = fxpar(headerc, 'CRPIX3')	; Reference channel 
zvel = fxpar(headerc, 'CRVAL3') ; Velocity of reference channel
lamda0=0.2110611405413D		; Rest wavelength of HI
f0 = 1420.40575177E6	        ; Rest frequency of HI
lamref  = ((zvel*lamda0) / ls) + lamda0     ; Wavelength of reference channel
fref = ls / lamref		; Frequency of reference channel
deltaf = 100E6 / 4096.0         ; Frequency resolution

; Parameters for sigma clipping, defaults are as for miriad
clip = 2.0			; Sigma level to clip to
nfi = 5				; Number of iterations

; Array for masking values during sigma clipping
weight = intarr(sizez)
weight(0:sizez-1) = 1

; Create an array to hold the channel numbers
channel = intarr(sizez)

for i = z1, z2-1 do begin 
   channel(i-z1) = i
endfor

; Make an array of velocity values based on channel numbers
velocity = -(ls*( ((channel - zref)*deltaf + fref - f0)/((channel - zref)*deltaf + fref) ))/1000.0

; Total number of spectra to examine
T = sizex*sizey

vmin = 0.0

for p = 1, sizex-2 do begin
  ; Percentage completeness of spectra examined
  print, 100*float(n)/float(T)

   for q = 1, sizey-2 do begin
      n = n + 1
     
      ; Check surrounding spectra of all cubes to make sure they're all good
      ; Tests indicate that is IS necessary to check the averaged cube and both pols,
      ; as in many cases only a few parts of the spectra may be bad
      ng = 0
      for i = p-1, p+1 do begin
         for j = q-1, q+1 do begin
            if (max(DATAC(i,j,z1:z2-1)) gt 0.0) and (max(DATAA(i,j,z1:z2-1)) gt 0.0) and (max(DATAB(i,j,z1:z2-1)) gt 0.0) then ng = ng + 1
         endfor
      endfor
   
      ; The raw spectrum along the required velocity range  
      value = DATAC(p,q,z1:z2-1)

      SW = 1.0

      ; Only operate when all spectra are good
      if (ng eq 9) then begin     

      ; The central pixel is already in memory - add the weighted surrounding pixels to
      ; it
      ; Loop is hardcoded to assume an integration area of 3x3
      ; This part of the code is shamelessly plagarised from mbspect
      for i = p-1, p+1 do begin
         for j = q-1, q+1 do begin
            dr=(float(i)-float(p))*cdelt1
            dd=(float(j)-float(q))*cdelt2
            w=exp(-2.0*alog10(4.0)*(dr^2+dd^2)/(beams*beams))
            if (i ne p) and (j ne q) then value = value + W*DATAC(i,j,z1:z2-1)
            SW = SW + w*w
         endfor
      endfor

      value = value / sqrt(SW)
      SW = 1.0
   
      nv = nv + 1
      
      ; Compute the 1st order baseline and subtract it
      ; Note that v refers to flux value, NOT velocity ! Should probably choose better variable names...
      vpoly = poly_fit(channel, value(0,0,0:sizez-1), 1)
      vapprox = vpoly(0) + vpoly(1)*channel

      vcorrected = value(0,0,0:sizez-1) - vapprox

      ; Compute velocity profile based on peak value, mask and estimate peak S/N

      ; Finds all values >= max value and puts them in an array of indicies
      peakchans = where(vcorrected ge max(vcorrected))
      ; Use the first value of this array - there should never be more than 1 - to set
      ; the masking range
      peakchan = peakchans(0)
      cst = peakchan - ctol/2.0		; Start channel
      cfi = peakchan + ctol/2.0		; Final channel
      ; Convert to integers, just in case
      cst = fix(cst)
      cfi = fix(cfi)

      ; Prevent masked range going outside of spectrum boundary
      if cst le 0 then begin
         cst = 0
         cfi = cst + ctol
      endif
         
      if cfi gt sizez-1 then begin
         cfi = sizez-1
         cst = cfi - ctol
      endif
 
      ; Create more temporary arrays without values in masked range
      goodtemp = fltarr(sizez-1-ctol)
      goodtempc= fltarr(sizez-1-ctol)

      j = 0
      for i = 0, sizez-1 do begin
         if (i lt cst) or (i gt cfi) then begin
            goodtemp(j) = vcorrected(0,0,i)
            goodtempc(j)= channel(i)
            j = j + 1
         endif
      endfor

      ; Fit and subtract a first-order baseline from masked spectrum
      ; Creates 2 arrays from fit - one to correct masked spectrum (to estimate rms),
      ; the other to correct the original spectrum to estimate velocity width.
      ; These arrays are identical in all but length !
      temppoly = poly_fit(goodtempc, goodtemp, 1)
      tempapprox = temppoly(0) + temppoly(1)*channel
      tempapprox2= temppoly(0) + temppoly(1)*goodtempc
    
      ; Corrected spectrum (with and without masking applied)
      tempcorrected = goodtemp - tempapprox2
      vcorrected2 = vcorrected(0,0,0:sizez-1) - tempapprox

      ; Calculate rms using corrected masked spectrum, with sigma clipping
      for ni=0, nfi do begin
         squares = 0D
         ;count = 0
         theend = size(tempcorrected, /DIMENSIONS)
         theend = double(theend(0))

         ; Apply baseline correction
         temppoly = poly_fit(goodtempc, tempcorrected, 1)
         tempapprox = temppoly(0) + temppoly(1)*goodtempc
         ;tempcorrected = tempcorrected - tempapprox
    
         for i = 0, theend-1 do begin 
            squares = squares + tempcorrected(i)*tempcorrected(i)
         endfor

         rmsc = sqrt(squares/theend)

         ; Convert to S/N
         SNarray = (tempcorrected - tempapprox) / rmsc
	
         ; Array of locations to delete
         delete = where(abs(SNarray) ge clip)

         ; Find the indices of points above clip and remove them completely
         if (max(abs(SNarray)) ge clip) then begin
            remove, delete, tempcorrected
            remove, delete, goodtempc
         endif


      endfor

      rmsc = sqrt(squares/theend)


      SN = max(vcorrected2(0,0,cst:cfi))/rmsc


      ; If S/N  sufficient, measure width minimiser at 50% of peak value
      ; Only examine masked/profile range
      ; Note the +cst is because IDL doesn't count as expected, see Notes.
      if SN ge S then begin
         maxchan = where(vcorrected2(0,0,cst:cfi) ge max(vcorrected2(0,0,cst:cfi))) + cst
         i = maxchan
         vchans = i
         half = max(vcorrected2(0,0,cst:cfi))/WL
         while (i ge 0) and (vcorrected2(i) gt half*0.95) do begin
            i = i - 1
            vchans = i
         endwhile

         i = maxchan
         vchanf = i
         while (i le sizez-1) and (vcorrected2(i) gt half*0.95) do begin
            i = i + 1
            vchanf = i
         endwhile

	 ; Because of how the while loops are evaluated, vchans and vchanf are the 
         ; first channels BELOW the 50% line.
         ; So the intersection points will occur between vchans and vchans+1, and also
         ; between vchanf and vchanf-1

         m = (vcorrected2(vchans) - vcorrected2(vchans+1)) / (vchans - (vchans+1))
         c = vcorrected2(vchans) - m*vchans

         ; m and c are arrays because IDL is daft
         min = (half - c(0)) / m(0)

         m = (vcorrected2(vchanf) - vcorrected2(vchanf-1)) / (vchanf - (vchanf-1))
         c = vcorrected2(vchanf) - m*vchanf

         max = (half - c(0)) / m(0)

         ; Can't properly evaluate if vchans or vchanf are at or exceed spectral edge
         if (vchans le 0.0) then min = 0.0
         if (vchanf ge sizez-1) then max = sizez-1

         ; Set intersection points to equivalent channel numbers
         ; Can't do channel(min) because that would round to integer value
         min = min + double(channel(0))
         max = max + double(channel(0))

         ; Get minimum and maximum velocities
         maxv = -(ls*( ((min - zref)*deltaf + fref - f0)/((min - zref)*deltaf + fref) ))/1000.0
	 minv = -(ls*( ((max - zref)*deltaf + fref - f0)/((max - zref)*deltaf + fref) ))/1000.0

         vmin = maxv - minv

       endif


        SNa = 0.0
        SNb = 0.0
        vmina = 0.0
        vminb = 0.0 

        ; If signal sufficient, check first polarisation
        ; Use same masking and profile range
        ; Effectively this is just a cut-and-paste job of the above code but using
        ; a different cube. No doubt this is hugely inefficient and shoddy coding, but
        ; oh well.
        if (SN ge S) and (vmin ge vw) then begin
           n1 = n1 + 1
           valuea = DATAA(p,q,z1:z2-1)

           for i = p-1, p+1 do begin
              for j = q-1, q+1 do begin
                 dr=(float(i)-float(p))*cdelt1
                 dd=(float(j)-float(q))*cdelt2
                 w=exp(-2.0*alog10(4.0)*(dr^2+dd^2)/(beams*beams))
                 if (i ne p) and (j ne q) then valuea = valuea + W*DATAA(i,j,z1:z2-1)
                 SW = SW + w*w
              endfor
           endfor

           valuea = valuea / sqrt(SW)
           SW = 1.0

           goodtemp = fltarr(sizez-1-ctol)
           goodtempc= fltarr(sizez-1-ctol)

           j = 0
           for i = 0, sizez-1 do begin
              if (i lt cst) or (i gt cfi) then begin
                 goodtemp(j) = valuea(0,0,i)
                 goodtempc(j)= channel(i)
                 j = j + 1
              endif
           endfor

           vpolya = poly_fit(goodtempc, goodtemp, 1)
           vapproxa = vpolya(0) + vpolya(1)*channel
           vapproxa2= vpolya(0) + vpolya(1)*goodtempc

           vcorrecteda = valuea - vapproxa    ; Corrected (without masking) for vw
           vcorrecteda2= goodtemp - vapproxa2 ; Corrected (with masking) for rms 


           weight(0:sizez-1) = 1     

           for ni=0, nfi do begin
             squares = 0D
             ;count = 0
             theend = size(vcorrecteda2, /DIMENSIONS)
             theend = double(theend(0))

             ; Apply baseline correction
             temppoly = poly_fit(goodtempc, vcorrecteda2, 1)
             tempapprox = temppoly(0) + temppoly(1)*goodtempc
             ;vcorrecteda2 = vcorrecteda2 - tempapprox
    
             for i = 0, theend-1 do begin 
                squares = squares + vcorrecteda2(i)*vcorrecteda2(i)
             endfor

             rmsa = sqrt(squares/theend)

             ; Convert to S/N
             SNarray = (vcorrecteda2 - tempapprox) / rmsa
	
             ; Array of locations to delete
             delete = where(abs(SNarray) ge clip)

             ; Find the indices of points above clip and remove them completely
             ;if delete ne -1 then begin
             if (max(abs(SNarray)) ge clip) then begin
               ;print, 'operating'
               remove, delete, vcorrecteda2
               remove, delete, goodtempc
            endif


           endfor


           rmsa = sqrt(squares/theend)

           SNa = max(vcorrecteda(0,0,cst:cfi)/rmsa)	; Peak S/N


           if SNa ge Si then begin
              maxchana = where(vcorrecteda(0,0,cst:cfi) ge max(vcorrecteda(0,0,cst:cfi))) + cst
              ;print, string(maxchan)+string(maxchana)+string(abs(maxchan-maxchana))
              i = maxchana
              vchansa = i
              half = max(vcorrecteda(0,0,cst:cfi))/WL
              while (i gt 0) and (vcorrecteda(i) gt half*0.95) do begin
                 i = i - 1
                 vchansa = i
              endwhile

              i = maxchana
              vchanfa = i
              while (i le sizez-1) and (vcorrecteda(i) gt half*0.95) do begin
                 i = i + 1
                 vchanfa = i
              endwhile

              m = (vcorrecteda(vchansa) - vcorrecteda(vchansa+1)) / (vchansa - (vchansa+1))
              c = vcorrecteda(vchansa) - m*vchansa

              mina = (half - c(0)) / m(0)

              m = (vcorrecteda(vchanfa) - vcorrecteda(vchanfa-1)) / (vchanfa - (vchanfa-1))
              c = vcorrecteda(vchanfa) - m*vchanfa
 
              maxa = (half - c(0)) / m(0)

              ; Can't properly evaluate if vchans or vchanf are at or exceed spectral edge
              if (vchansa le 0.0) then mina = 0.0
              if (vchanfa ge sizez-1) then maxa = sizez-1

              ; Set intersection points to equivalent channel numbers
              ; Can't do channel(min) because that would round to integer value
              mina = mina + double(channel(0))
              maxa = maxa + double(channel(0))

              ; Get minimum and maximum velocities
              maxva = -(ls*( ((min - zref)*deltaf + fref - f0)/((min - zref)*deltaf + fref) ))/1000.0
	      minva = -(ls*( ((max - zref)*deltaf + fref - f0)/((max - zref)*deltaf + fref) ))/1000.0

              vmina = maxva - minva

          
           endif
     


        ; If this signal sufficient then check second polarisation
        if (SNa ge Si) and (vmina ge vw) then begin
           n2 = n2 + 1
           valueb = DATAB(p,q,z1:z2-1)

           SW = 1.0

           for i = p-1, p+1 do begin
              for j = q-1, q+1 do begin
                 dr=(float(i)-float(p))*cdelt1
                 dd=(float(j)-float(q))*cdelt2
                 w=exp(-2.0*alog10(4.0)*(dr^2+dd^2)/(beams*beams))
                 if (i ne p) and (j ne q) then valueb = valueb + W*DATAB(i,j,z1:z2-1)
                 SW = SW + w*w
              endfor
           endfor

           valueb = valueb / sqrt(SW)
           SW = 1.0
          

           goodtemp = fltarr(sizez-1-ctol)
           goodtempc= fltarr(sizez-1-ctol)

           j = 0
           for i = 0, sizez-1 do begin
              if (i lt cst) or (i gt cfi) then begin
                 goodtemp(j) = valueb(0,0,i)
                 goodtempc(j)= channel(i)
                 j = j + 1
              endif
           endfor

           vpolyb = poly_fit(goodtempc, goodtemp, 1)
           vapproxb = vpolyb(0) + vpolyb(1)*channel
           vapproxb2= vpolyb(0) + vpolyb(1)*goodtempc

           vcorrectedb = valueb - vapproxb    ; Corrected (without masking) for vw
           vcorrectedb2= goodtemp - vapproxb2 ; Corrected (with masking) for rms 
         
  
           ;weight(0:sizez-1) = 1 

           for ni=0, nfi do begin
              squares = 0D
              ;count = 0
              theend = size(vcorrectedb2, /DIMENSIONS)
              theend = double(theend(0))

              ; Apply baseline correction
              temppoly = poly_fit(goodtempc, vcorrectedb2, 1)
              tempapprox = temppoly(0) + temppoly(1)*goodtempc
              ;tempcorrected = vcorrectedb2 - tempapprox
    
              for i = 0, theend-1 do begin 
                 squares = squares + vcorrectedb2(i)*vcorrectedb2(i)
              endfor

              rmsb = sqrt(squares/theend)

              ; Convert to S/N
              SNarray = (vcorrectedb2 - tempapprox) / rmsb
	
              ; Array of locations to delete
              delete = where(abs(SNarray) ge clip)

              ; Find the indices of points above clip and remove them completely
              ;if delete ne -1 then begin
              if (max(abs(SNarray)) ge clip) then begin
                 remove, delete, vcorrectedb2
                 remove, delete, goodtempc
              endif


            endfor

            rmsb = sqrt(squares/theend)
 
            SNb = max(vcorrectedb(0,0,cst:cfi))/rmsb

            theend = size(vcorrectedb2, /DIMENSIONS)
            theend = double(theend(0))

    
           if SNb ge Si then begin
              maxchanb = where(vcorrectedb(0,0,cst:cfi) ge max(vcorrectedb(0,0,cst:cfi))) + cst
              i = maxchanb
              vchansb = i
              half = max(vcorrectedb(0,0,cst:cfi))/WL
              while (i gt 0) and (vcorrectedb(i) gt half*0.95) do begin
                 i = i - 1
                 vchansb = i
              endwhile

              i = maxchanb
              vchanfb = i
              while (i le sizez-1) and (vcorrectedb(i) gt half*0.95) do begin
                 i = i + 1
                 vchanfb = i
              endwhile

              m = (vcorrectedb(vchansb) - vcorrectedb(vchansb+1)) / (vchansb - (vchansb+1))
              c = vcorrectedb(vchansb) - m*vchansb

              minb = (half - c(0)) / m(0)

              m = (vcorrectedb(vchanfb) - vcorrectedb(vchanfb-1)) / (vchanfb-1 - (vchanfb-1))
              c = vcorrectedb(vchanfb) - m*vchanfb

              maxb = (half - c(0)) / m(0)

              ; Can't properly evaluate if vchans or vchanf are at or exceed spectral edge
              if (vchansb le 0.0) then minb = 0.0
              if (vchanfb ge sizez-1) then maxb = sizez-1

              ; Set intersection points to equivalent channel numbers
              ; Can't do channel(min) because that would round to integer value
              minb = minb + double(channel(0))
              maxb = maxb + double(channel(0))

              ; Get minimum and maximum velocities
              maxvb = -(ls*( ((min - zref)*deltaf + fref - f0)/((min - zref)*deltaf + fref) ))/1000.0
	      minvb = -(ls*( ((max - zref)*deltaf + fref - f0)/((max - zref)*deltaf + fref) ))/1000.0

              vminb = maxvb - minvb

             if (vminb ge vw) then begin
                 n3 = n3 + 1

                 xyad, HEADERC, p, q, ra, dec		; Convert pixel to sky
                 coords=adstring(ra, dec)               ; Format into sexigesimal
                 sc = strsplit(coords, ' ', /extract)
                 coords=' '+sc[0]+':'+sc[1]+':'+sc[2]+' '+sc[3]+':'+sc[4]+':'+sc[5]

                 openw, lun, 'Output.txt', /get_lun, width=200, /APPEND
                 printf, lun, string(p)+string(q)+string(peakchan+z1)+coords+string(velocity(peakchan))+string(SN)+string(vmin)+string(cst+z1)+string(cfi+z1)
                 free_lun, lun
              endif
           endif
         endif

          
        endif

        endif


   endfor
endfor

print, 'Total number of spectra examined :', n
print, 'Total number of valid spectra :', nv
print, 'Number of candidates from averaged cube :', n1
print, 'Number left after examining polarisation A:', n2
print, 'Number matched in both polarisations :', n3

; STAGE 2 - Search the output file for matches
; Runs a batch file that calls starlink packages to process the ouput data (same as 
; Topcat uses)

print, 'Reducing the catalogue to highest S/N pixels in each group...'

spawn, '/home/virgo/spxrt/AGES7448Stuff/CHELL.cmd'

print, 'Final number of sources :'

spawn, 'wc -l Catalogue.txt'

; Next, use the catalogue file to go back and re-extract the brightest matched sources
; and plot them
; Size have already computed the important parameters there is no need to remeasure
; everything, only to correct the baseline and plot stuff.

print, 'Plotting spectra...'

readcol, 'Catalogue.txt', x, y, z, ra, dec, velocity2, SN, vmin, mask1, mask2, GroupID, GroupSize, format='LL, LL, LL, A, A, F, F, F, LL, LL, LL, LL'
length = size(x, /DIMENSIONS)
l = length(0)

for i=0, l-1 do begin
   value=DATAC(x(i),y(i),z1:z2-1)

   ; Create arrays with masked channels for estimating the baseline
   tempflux = fltarr(sizez-1-ctol)
   tempchan = fltarr(sizez-1-ctol)

   j = 0
   for k = 0, sizez-1 do begin
      if (k lt (mask1(i)-z1)) or (k gt (mask2(i)-z1)) then begin
         tempflux(j) = value(0,0,k)
         tempchan(j)= channel(k)
         j = j + 1
      endif
   endfor
   
   fpoly = poly_fit(tempchan, tempflux(0:sizez-2-ctol), 1)
   fapprox = fpoly(0) + fpoly(1)*channel

   fcorrected = value(0,0,0:sizez-1) - fapprox
   xyad, HEADERC, x(i), y(i), ra, dec
   coords=adstring(ra, dec)
   sc = strsplit(coords, ' ', /extract)
   coords=' '+sc[0]+':'+sc[1]+':'+sc[2]+' '+sc[3]+':'+sc[4]+':'+sc[5]

   SNs=string(SN(i))
   vws=string(vmin(i))
   fulltitle = 'RA, Dec = '+coords+' SN = '+SNs+' W50 min = '+vws
   propertitle = strcompress(fulltitle)

   
   plot, velocity, fcorrected(0,0,0:sizez-1), xrange=[velocity(sizez-1),velocity(0)], title=propertitle ; Calibrated spectrum

   oplot, [velocity(mask1(i)-z1),velocity(mask1(i)-z1)], [-1,1];, color=-500                   ;Masked range
   oplot, [velocity(mask2(i)-z1),velocity(mask2(i)-z1)], [-1,1];, color=-500

   ;if (x(i) eq 187) and (y(i) eq 59) then begin
   ;   openw, lun, 'SourceTest.txt'
   ;   for i = 0, sizez-1 do begin
   ;      printf, lun, velocity(i), fcorrected(0,0,i)
   ;   endfor
   ;   free_lun, lun
   ;endif


endfor

print, 'DONE !'

spawn, 'date'

device, /close

end
